package gelwood.crossword;

import java.io.FileInputStream;
import java.util.ArrayList;
import java.util.Iterator;

public class Solver {

	private Grid grid;
	private Dictionary dictionary;
	
	private ArrayList<Decision> decisions;
	
	private Word currentSlot;
	
	public Solver(String gridPath, String dictPath){
		this.grid = new Grid(this.fileToString(gridPath));
		this.dictionary = new Dictionary(this.fileToString(dictPath));
		this.decisions = new ArrayList<Decision>();
		
		this.grid.setupSubDictionaries(this.dictionary);
		
		
		
		if (this.grid.hasMoreSlots()) {
			this.grid.sortWordsByWeight();
			
			this.currentSlot = this.grid.getFirstSlot();
			
			this.printState("START");
			
			
		} else {
			System.out.println("No Slots in Grid");
		}
		
		//if any of the slots, have no words in their dictionary, fail
		
		ArrayList<Word> theseSlots = this.grid.getSortedWordSlots();
		Iterator<Word> theseSlotsIT = theseSlots.iterator();
		
		ArrayList<Integer> lengthsInDict = this.dictionary.getWordLengths();
		
		
		boolean solvable = true;
		
		while(theseSlotsIT.hasNext()){
			Word myWord = theseSlotsIT.next();
			int myLength = myWord.getLength();
			
			Iterator<Integer> mylengths = lengthsInDict.iterator();
			
			solvable = solvable && this.containsInt(mylengths, myLength);
			
		}
		
		if (solvable) {
			this.decide();
		} else {
			this.fail();
		}
		
	}
	

	
	private void fail(){
		this.printState("FAIL");
	}
	
	//this method chooses a random word that fits into the current slot
	//
	private void decide(){
		Dictionary currentD = this.currentSlot.subDictionary;
		
		//if the decisions in the grid have created slots with no possible
		//solution, back track
		if (grid.isContradiction()) {
			this.printState("DECIDE");
			
			this.backtrack();
		//otherwise
		//choose a word for the current slot
			
		} else {
			String newWord = currentD.removeRandomWord();
			
			//String newWord = this.grid.chooseBestWord(currentD);
			
			if (this.grid.isUsed(newWord)) {
				this.decide();
			} else {
				ArrayList<Word> intersections = this.grid.getIntersectingWords(this.currentSlot);
			
				this.setWord(newWord, intersections);
			
				this.printState("DECIDE");
			
				if (intersections.size() == 0) {
					this.restart();
				} else {
					this.chooseNextSlot(intersections);
				}	
			}
		}	
	}
	
	//this method chooses which slot to try to solve next
	private void chooseNextSlot(ArrayList<Word> intersections){
		this.currentSlot = this.chooseHelp(intersections);
		
		System.out.println("CURRENT SLOT");
		this.currentSlot.print();
		
		this.currentSlot.updateDictionary();
		
		Dictionary currentD = this.currentSlot.subDictionary;
		
		if (currentD.getSize() > 0) {
			this.decide();
		} else {
			this.backtrack();
		}
	}
	
	private Word chooseHelp(ArrayList<Word> intersections){
		Iterator<Word> intersectionsIt = intersections.iterator();
		
		Word myWord = intersectionsIt.next();
		
		int i = myWord.getHeuristic();
		
		return chooseHelp(i, myWord, intersectionsIt);
	}
	
	private Word chooseHelp(int i, Word myWord, Iterator<Word> intersections){
		if (intersections.hasNext()) {
			Word currentWord = intersections.next();
			
			int heuristic = currentWord.getHeuristic();
			
			if (heuristic < i) {
				return chooseHelp(heuristic, currentWord, intersections);
			} else {
				return chooseHelp(i, myWord, intersections);
			}
			
		}
		else {
			return myWord;
		}
	}
	
	private void restart(){
		printState("Restart");
		
		ArrayList<Word> slots = this.grid.getSortedWordSlots();
		if (slots.size() == 0) {
			this.succeed();
		} else {
			this.currentSlot = this.grid.getFirstSlot(); 
			
			this.decide();
		}
	}

	private void succeed(){
		
		printState("Succeed");
	}
	
	private void backtrack(){

		this.printState("Backtrack");
		System.out.println();
		
		Iterator<Decision> myIT = decisions.iterator();
		while(myIT.hasNext()) {
			Decision myDecision = myIT.next();
			
			myDecision.print();
		}
		
		
		this.resetWord();
		
	}
	
	private void resetWord(){
		int dIndex = this.decisions.size();
		Decision lastDecision = this.decisions.remove(dIndex - 1);
		
		lastDecision.print();
		
		this.currentSlot = lastDecision.getSlot();
		
		this.grid.resetSlot(lastDecision);
		
		lastDecision.resetIntersections();
		
		this.decide();
	}
	
	private void setWord(String newWord, ArrayList<Word> intersections){
		Decision myDecision = new Decision(newWord, this.currentSlot);
		
		myDecision.setIntersections(intersections);
		
		this.grid.setSlot(myDecision);
		
		this.decisions.add(myDecision);
		
	}
	
	private boolean containsInt(Iterator<Integer> mylengths, int length){
		if (mylengths.hasNext()){
			Integer x = mylengths.next();
			if (x == length) {
				return true;
			} else {
				return this.containsInt(mylengths, length);
			}
		} else {
			return false;
		}
	}
	
	private void printState(String state){
		System.out.println("\n\n\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
		System.out.println(state);
		System.out.println("");
		this.grid.print();
		System.out.println("Current Slot: ");
		this.currentSlot.print();
		System.out.println("Decisions : " + decisions.size());
		System.out.println("\n\n");
	}
	
	private String fileToString(String path){
		byte[] myBytes = null;
		
		try {
			FileInputStream myFile = new FileInputStream(path);
			myBytes = new byte[myFile.available()];
			myFile.read(myBytes);
			myFile.close();
		} catch (Exception e) {
			System.out.println("Exception in Grid.fileToString: " + e);
		}
		
		return new String(myBytes);
	}
}
